/*
------------------------------------------------------------
   Kobo Deluxe - An enhanced SDL port of XKobo
------------------------------------------------------------
 * Copyright (C) 1995, 1996, Akira Higuchi
 * Copyright (C) 2001, David Olofson
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>

#include "manage.h"
#include "kobo.h"
#include "gamectl.h"

int gamecontrol_t::space;
int gamecontrol_t::left;
int gamecontrol_t::up;
int gamecontrol_t::down;
int gamecontrol_t::right;
int gamecontrol_t::dl;
int gamecontrol_t::dr;
int gamecontrol_t::ul;
int gamecontrol_t::ur;
int gamecontrol_t::shot;
int gamecontrol_t::direction;
int gamecontrol_t::movekey_pressed;

void gamecontrol_t::init()
{
	direction = 1;
	left = 0;
	right = 0;
	up = 0;
	down = 0;
	shot = 0;
	movekey_pressed = 0;
}

void gamecontrol_t::clear()
{
	direction = 1;
}


int gamecontrol_t::map(SDLKey sym)
{
/*
FIXME: This should be replaced by a configurable mapping system.
*/
	switch(sym)
	{
	  case KEY_KP_LEFT:
	  case SDLK_LEFT:
		return BTN_LEFT;
	  case KEY_KP_RIGHT:
	  case SDLK_RIGHT:
		return BTN_RIGHT;
	  case KEY_KP_UP:
	  case SDLK_UP:
		return BTN_UP;
	  case KEY_KP_DOWN:
	  case SDLK_DOWN:
		return BTN_DOWN;
	  case KEY_KP_DL:
	  case SDLK_END:
		return BTN_DL;
	  case KEY_KP_DR:
	  case SDLK_PAGEDOWN:
		return BTN_DR;
	  case KEY_KP_UL:
	  case SDLK_HOME:
		return BTN_UL;
	  case KEY_KP_UR:
	  case SDLK_PAGEUP:
		return BTN_UR;
	  case SDLK_LSHIFT:
	  case SDLK_RSHIFT:
	  case SDLK_LCTRL:
	  case SDLK_RCTRL:
		return BTN_FIRE;
	  case SDLK_ESCAPE:
		return BTN_EXIT;
	  case SDLK_SPACE:
		return BTN_START;
	  case SDLK_KP_ENTER:
	  case SDLK_RETURN:
		return BTN_SELECT;
	  case SDLK_KP_PLUS:
		return BTN_INC;
	  case SDLK_KP_MINUS:
		return BTN_DEC;
	  case SDLK_y:
		return BTN_YES;
	  case SDLK_n:
		return BTN_NO;
	  case SDLK_BACKSPACE:
		return BTN_REGEN;
	  default:
		return -1;
	}
}


void gamecontrol_t::press(int k)
{
	switch(k)
	{
	  case BTN_LEFT:
		left = 1;
		break;
	  case BTN_RIGHT:
		right = 1;
		break;
	  case BTN_UP:
		up = 1;
		break;
	  case BTN_DOWN:
		down = 1;
		break;
	  case BTN_UL:
		ul = 1;
		break;
	  case BTN_UR:
		ur = 1;
		break;
	  case BTN_DL:
		dl = 1;
		break;
	  case BTN_DR:
		dr = 1;
		break;
	  case BTN_FIRE:
		shot = 1;
		break;
	}
	gamecontrol_t::change();
}

void gamecontrol_t::release(int k)
{
	switch(k)
	{
	  case BTN_LEFT:
		left = 0;
		break;
	  case BTN_RIGHT:
		right = 0;
		break;
	  case BTN_UP:
		up = 0;
		break;
	  case BTN_DOWN:
		down = 0;
		break;
	  case BTN_UL:
		ul = 0;
		break;
	  case BTN_UR:
		ur = 0;
		break;
	  case BTN_DL:
		dl = 0;
		break;
	  case BTN_DR:
		dr = 0;
		break;
	  case BTN_FIRE:
		shot = 0;
		break;
	}
	gamecontrol_t::change();
}

void gamecontrol_t::change_up(unsigned long r)
{
	up = (r == 0) ? 1 : 0;
	gamecontrol_t::change();
}

void gamecontrol_t::change_down(unsigned long r)
{
	down = (r == 0) ? 1 : 0;
	gamecontrol_t::change();
}

void gamecontrol_t::change_left(unsigned long r)
{
	left = (r == 0) ? 1 : 0;
	gamecontrol_t::change();
}

void gamecontrol_t::change_right(unsigned long r)
{
	right = (r == 0) ? 1 : 0;
	gamecontrol_t::change();
}

void gamecontrol_t::change_shot(unsigned long r)
{
	shot = (r == 0) ? 1 : 0;
	gamecontrol_t::change();
}

void gamecontrol_t::mouse_press(int n)
{
	if(n == 1)
		shot = 1;
//	else if(n == 3)
//		manage.key_down(KEY_START);
	gamecontrol_t::change();
}

void gamecontrol_t::mouse_release(int n)
{
	if(n == 1)
		shot = 0;
	gamecontrol_t::change();
}

void gamecontrol_t::mouse_position(int h, int v)
{
	if(h > 0)
	{
		if(v > 0)
		{
			if(h > (v << 1))
				direction = 3;
			else if(v > (h << 1))
				direction = 5;
			else
				direction = 4;
		}
		else if(v < 0)
		{
			if(h > ((-v) << 1))
				direction = 3;
			else if((-v) > (h << 1))
				direction = 1;
			else
				direction = 2;
		}
		else
			direction = 3;
	}
	else if(h < 0)
	{
		if(v > 0)
		{
			if((-h) > (v << 1))
				direction = 7;
			else if(v > ((-h) << 1))
				direction = 5;
			else
				direction = 6;
		}
		else if(v < 0)
		{
			if((-h) > ((-v) << 1))
				direction = 7;
			else if((-v) > ((-h) << 1))
				direction = 1;
			else
				direction = 8;
		}
		else
			direction = 7;
	}
	else if(v > 0)
	{
		direction = 5;
	}
	else
	{
		direction = 1;
	}
}


void gamecontrol_t::change()
{
	int lr = left - right + ul - ur + dl - dr;
	int ud = up - down + ul + ur - dl - dr;
	if(lr > 0)
	{
		if(ud > 0)
			direction = 8;
		else if(ud < 0)
			direction = 6;
		else
			direction = 7;
	}
	else if(lr < 0)
	{
		if(ud > 0)
			direction = 2;
		else if(ud < 0)
			direction = 4;
		else
			direction = 3;
	}
	else
	{
		if(ud > 0)
			direction = 1;
		else if(ud < 0)
			direction = 5;
	}
	movekey_pressed = left | right | up | down;
	mouse_x = 0;
	mouse_y = 0;
}
