/*
------------------------------------------------------------
   Kobo Deluxe - An enhanced SDL port of XKobo
------------------------------------------------------------
 * Copyright (C) 1995, 1996, Akira Higuchi
 * Copyright (C) 2001, David Olofson
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#undef	DEBUG_OUT

extern "C"
{
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
}

#include "config.h"
#include "kobo.h"
#include "states.h"
#include "screen.h"
#include "manage.h"
#include "score.h"
#include "radar.h"
#include "gamectl.h"
#include "random.h"
#include "version.h"
#include "sounds.h"
#include "bgmfile.h"
#include "options.h"
#include "myship.h"


skobo_gfxengine_t gengine;
filemapper_t fmap;
prefs_t prefs;

SDL_Joystick *joystick = NULL;
int js_lr = DEFAULT_JOY_LR;
int js_ud = DEFAULT_JOY_UD;
int js_fire = DEFAULT_JOY_FIRE;
int js_start = DEFAULT_JOY_START;

window_t wbase;
window_t wchip;
window_t wradar;
//window_t wscore;

display_t dhigh;
display_t dscore;
display_t dstage;
display_t dships;

int exit_game = 0;
int exit_game_fast = 0;

int debug_mode = 0;
int noframe = 0;

int show_fps = 0;
static int fps_count = 0;
static int fps_starttime = 0;
display_t dfps;

int intscale = 1;
int xoffs = 0;
int yoffs = 0;


#if 0
void motion_callback(win &)
{
	mouse_x = win::xevt.xmotion.x;
	mouse_y = win::xevt.xmotion.y;
	mouse_x -= (WSIZE >> (1 - scale_log2));
	mouse_y -= (WSIZE >> (1 - scale_log2));
	if ((mouse_x == 0) && (mouse_y == 0))
		return;
	gamecontrol.mouse_position(mouse_x, mouse_y);
}

void leave_callback(win &)
{
	mouse_x = 0;
	mouse_y = 0;
}

void focus_in_callback(win &)
{
	gamecontrol.clear();
}

void focus_out_callback(win &)
{
	gamecontrol.clear();
}

void button_press_callback(win &)
{
	gamecontrol.mouse_press(win::xevt.xbutton.button);
}

void button_release_callback(win &)
{
	gamecontrol.mouse_release(win::xevt.xbutton.button);
}

void client_message_callback(win &)
{
	Atom wm_protocols, wm_delete_window;
	wm_protocols = XInternAtom(win::disp, "WM_PROTOCOLS", False);
	wm_delete_window =
		XInternAtom(win::disp, "WM_DELETE_WINDOW", False);
	if ((win::xevt.xclient.message_type == wm_protocols)
		&& (win::xevt.xclient.data.l[0] == wm_delete_window))
	{
		_manage::exit_key();
		_manage::exit_key();
	}
}
#endif


#define	NIBBLE_W	4
#define	NIBBLE_H	4
#define	NIBBLE_TILES	((SCREEN_WIDTH/NIBBLE_W+1)*(SCREEN_HEIGHT/NIBBLE_H+1))

/*
 * Tools:
 *	-1: Random tool
 *	0:  Black 4x4 square
 *	1+: Brushes from "brushes.png";
 *		1: Soft black 16 pixel round brush
 *		2: (4/8)x4 tilable romboid brush
 *		3: Tilable garbage brush
 *		4: Vertical "D" brush
 */
void nibble(int tool = -1)
{
	int i;
	int x[NIBBLE_TILES];
	int y[NIBBLE_TILES];

	if(tool < 0)
		tool = rand_num.get(20) % 5;

	/* Clear */
	for(i = 0; i < NIBBLE_TILES; ++i)
		x[i] = y[i] = -1;

	/* Fill in */
	int index = 0;
	int xx = -NIBBLE_W/2;
	int yy = -NIBBLE_H/2;
	for(i = 0; i < NIBBLE_TILES; ++i)
	{
		index = rand_num.get() % NIBBLE_TILES;
		if(index >= NIBBLE_TILES)
			index = 0;
		while(x[index] != -1)
			if(++index >= NIBBLE_TILES)
				index = 0;
		x[index] = xx;
		y[index] = yy;
		xx += NIBBLE_W;
		if(xx >= SCREEN_WIDTH + NIBBLE_W/2)
		{
			xx = -NIBBLE_W/2;
			yy += NIBBLE_H;
		}
	}

	/* Clear 4000 tiles/second, until all are done. */
//	wbase.foreground(wbase.alloc_color(0xffff,0xffff,0xffff));
//	wbase.fillrect(0, 0, wbase.width(), wbase.height());
	wbase.foreground(wbase.alloc_color(0,0,0));
	index = 0;
	int t = SDL_GetTicks();
	while(index < NIBBLE_TILES)
	{
		int nt = SDL_GetTicks();
		int dt = nt - t;
		t = nt;
		for(int i = 0; i < dt * 4; ++i)
		{
			if(index >= NIBBLE_TILES)
				break;
			switch(tool)
			{
			  case 0:
				wbase.fillrect(x[index] + NIBBLE_W/2,
						y[index] + NIBBLE_H/2,
						NIBBLE_W, NIBBLE_H);
				break;
			  default:
				wbase.sprite(x[index]-8 + NIBBLE_W/2,
						y[index]-8 + NIBBLE_H/2,
						B_BRUSHES, tool-1);
				break;
			}
			++index;
		}
		gengine.invalidate();
		gengine.flip();
	}
	wbase.fillrect(0, 0, wbase.width(), wbase.height());
	SDL_Delay(200);
}




void put_usage()
{
	printf("\nKobo Deluxe %s\n", KOBO_VERSION);
	printf("Usage: kobodl -hiscores\n");
	printf("       kobodl -highscores\n");
	printf("       kobodl -showcfg\n");
	printf("       kobodl [-noframe]\n");
	printf("              [-fps]\n");
	printf("              [-debug]\n");
	printf("              [-[no]fullscreen]\n");
	printf("              [-buffer <n>] (default: 0)\n");
	printf("              [-depth <n>] (default: 0)\n");
	printf("              [-[no]nofilter]\n");
	printf("              [-[no]cheat]\n");
	printf("              [-[no]indicator]\n");
	printf("              [-wait <n>] (default: %d)\n", WAIT_MSEC);
	printf("              [-[no]nosound]\n");
#ifdef OSS_SUPPORT
	printf("              [-[no]oss]\n");
#endif
	printf("              [-latency <n>] (default: 50)\n");
	printf("              [-vol <n>] (default: 100)\n");
	printf("              [-reverb <n>] (default: 100)\n");
	printf("              [-threshold <n>] (default: 100)\n");
	printf("              [-release <n>] (default: 100)\n");
	printf("              [-bgm <bgm index>]\n");
	printf("              [-files <kobo data dir path>]\n");
	printf("              [-gfx <gfx dir path>]\n");
	printf("              [-scores <scores dir path>]\n");
	printf("              [-[no]joystick]\n");
	printf("\n");
	exit(1);
}


void setup_dirs(char *xpath)
{
	fmap.exepath(xpath);

	/*
	 * Graphics data
	 */
	fmap.addpath("DATA", KOBO_DATA_DIR);
	fmap.addpath("GFX", "DATA>>gfx");
	/* Current dir */
	fmap.addpath("GFX", "./gfx");
	/* Current dir; from within the build tree */
	fmap.addpath("GFX", "./data");

	/*
	 * Score files (user and global)
	 */
	fmap.addpath("SCORES", KOBO_SCORE_DIR);
	/* 'scores' in current dir (For importing scores, perhaps...) */
	fmap.addpath("SCORES", "./scores");

	/*
	 * Configuration files
	 */
	fmap.addpath("CONFIG", KOBO_CONFIG_DIR);
	/* System local */
	fmap.addpath("CONFIG", SYSCONF_DIR);
	/* In current dir (last resort) */
	fmap.addpath("CONFIG", "./");
}


void add_dirs(prefs_t *p)
{
	char buf[300];
	if(p->dir[0])
	{
		char *upath = fmap.sys2unix(p->dir);
		snprintf(buf, 300, "%s/gfx", upath);
		fmap.addpath("GFX", buf, 1);
		snprintf(buf, 300, "%s/scores", upath);
		fmap.addpath("SCORES", buf, 1);
		snprintf(buf, 300, "%s", upath);
		fmap.addpath("CONFIG", buf, 0);
	}

	if(p->gfxdir[0])
		fmap.addpath("GFX", fmap.sys2unix(p->gfxdir), 1);

	if(p->scoredir[0])
		fmap.addpath("SCORES", fmap.sys2unix(p->scoredir), 1);
}


int init_display(prefs_t *p)
{
	gengine.title("Kobo Deluxe v" VERSION, "kobodl");
	gengine.driver((gfx_drivers_t)p->videodriver);
	gengine.scale(p->width/SCREEN_WIDTH, p->height/SCREEN_HEIGHT);
	intscale = p->width/SCREEN_WIDTH;
	if(p->videodriver == GFX_DRIVER_SDL2D)
	{
		xoffs = (p->width - intscale*SCREEN_WIDTH) / (intscale*2);
		yoffs = (p->height - intscale*SCREEN_HEIGHT) / (intscale*2);
	}
	if(p->fullscreen)
		gengine.size(p->width, p->height);
	else
	{
		int border = 4/intscale*intscale;
		xoffs += border/intscale;
		yoffs += border/intscale;
		gengine.size(p->width+border*2, p->height+border*2);
	}
	gengine.mode(0, p->fullscreen);
	gengine.buffering((gfx_buffermodes_t)(p->buffermode), GFX_SYNC_FLIP);
	gengine.cursor(0);
	gengine.period(p->wait_msec);
	gengine.interpolation(p->filter);
	gengine.scroll_ratio(LAYER_OVERLAY, 0.0, 0.0);
	gengine.scroll_ratio(LAYER_PLAYER, 1.0, 1.0);
	gengine.scroll_ratio(LAYER_ENEMIES, 1.0, 1.0);
	gengine.scroll_ratio(LAYER_BASES, 1.0, 1.0);
	gengine.scroll_ratio(LAYER_GROUND, 1.0, 1.0);
	gengine.scroll_ratio(LAYER_STARS, 0.5, 0.5);
	gengine.wrap(MAP_SIZEX*CHIP_SIZEX, MAP_SIZEY*CHIP_SIZEY);
	wbase.init(&gengine);
	wbase.place(xoffs, yoffs, SCREEN_WIDTH, SCREEN_HEIGHT);
	return gengine.open(2048);
}


void progress(int percent)
{
	int x, y, w, h;
	x = 0;
	w = wbase.width() * percent / 100;
	if(w < 4)
		return;

	h = 16;
	y = wbase.height()-h;

	SDL_Rect r;
	r.x = x;
	r.y = y;
	r.w = w;
	r.h = h;

	wbase.foreground(wbase.alloc_color(0,0,0x9999));
	wbase.rectangle(x,y,w,h);

	++x; ++y; w-=2; h-=2;
	wbase.foreground(wbase.alloc_color(0,0,0xcccc));
	wbase.rectangle(x,y,w,h);

	++x; ++y; w-=2; h-=2;
	wbase.foreground(wbase.alloc_color(0,0,0xffff));
	wbase.fillrect(x,y,w,h);

	wbase.invalidate(&r);
	gengine.flip();
}


#define	TUBE_LENGTH	52


int load_graphics(prefs_t *p)
{
	int i;
	const char *fn;
	SDL_Rect r;
	gengine.colorkey(65535, 0, 0);
	gengine.scalemode((gfx_scalemodes_t)p->scalemode);

	fn = fmap.get("GFX>>loading.png");
	if(!fn || gengine.loadimage(B_LOADING, fn) < 0)
		return -50;
	for(i = 0; i < 2; ++i)
	{
		wbase.background(wbase.alloc_color(0,0,0));
		wbase.clear();
		wbase.sprite((wbase.width()-158)/2, 60, B_LOADING, 0);
		gengine.invalidate();
		gengine.flip();
	}

	fn = fmap.get("GFX>>tiles.pcx");
	if(!fn || gengine.loadtiles(B_TILES, 16, 16, fn) < 0)
		return -1;
	progress(10);
	fn = fmap.get("GFX>>sprites.pcx");
	if(!fn || gengine.loadsprites(B_SPRITES, 16, 16, fn) < 0)
		return -2;
	progress(20);
	fn = fmap.get("GFX>>bullets.pcx");
	if(!fn || gengine.loadsprites(B_BULLETS, 8, 8, fn) < 0)
		return -3;
	progress(22);
	fn = fmap.get("GFX>>bigship.pcx");
	if(!fn || gengine.loadsprites(B_BIGSHIP, 32, 32, fn) < 0)
		return -4;
	progress(25);

	fn = fmap.get("GFX>>screen.png");
	if(!fn || gengine.loadimage(B_SCREEN, fn) < 0)
		return -5;
	progress(55);
	r.x = (8+MARGIN) * intscale;
	r.y = MARGIN * intscale;
	r.w = r.h = WSIZE * intscale;
	if(gengine.loadrect(B_FRAME, B_SCREEN, 0, &r) < 0)
		return -6;
	progress(57);

	fn = fmap.get("GFX>>logo3.png");
	if(!fn || gengine.loadimage(B_LOGO, fn) < 0)
		return -7;
	progress(70);

	fn = fmap.get("GFX>>font2b.png");
	if(!fn || gengine.loadfont(B_NORMAL_FONT, fn) < 0)
		return -8;
	progress(75);

	fn = fmap.get("GFX>>font3b.png");
	if(!fn || gengine.loadfont(B_BIG_FONT, fn) < 0)
		return -9;
	progress(80);

	fn = fmap.get("GFX>>counter.png");
	if(!fn || gengine.loadfont(B_COUNTER_FONT, fn) < 0)
		return -10;
	progress(90);

	fn = fmap.get("GFX>>brushes.png");
	if(!fn || gengine.loadsprites(B_BRUSHES, 16, 16, fn) < 0)
		return -11;
	progress(95);

	fn = fmap.get("GFX>>tube1.png");
	if(!fn || gengine.loadimage(B_TEMP, fn) < 0)
		return -12;
	r.x = (60-TUBE_LENGTH) * intscale / 2;
	r.y = 0;
	r.w = r.h = TUBE_LENGTH * intscale;
	if(gengine.loadrect(B_BIG_TUBE, B_TEMP, 0, &r) < 0)
		return -13;
	gengine.unload(B_TEMP);
	progress(97);

	fn = fmap.get("GFX>>tube2.png");
	if(!fn || gengine.loadimage(B_TEMP, fn) < 0)
		return -12;
	r.x = (60-TUBE_LENGTH) * intscale / 2;
	r.y = 0;
	r.w = r.h = TUBE_LENGTH * intscale;
	if(gengine.loadrect(B_SMALL_TUBE, B_TEMP, 0, &r) < 0)
		return -13;
	gengine.unload(B_TEMP);
	progress(100);

	nibble();
	return 0;
}


void build_screen()
{
	wbase.font(B_NORMAL_FONT);
	wbase.foreground(wbase.alloc_color(65535, 65535, 65535));
	wbase.background(wbase.alloc_color(0x4600, 0x5800, 0x6500));
	wbase.sprite(0, 0, B_SCREEN, 0);

	wchip.init(&gengine);
	wchip.place(xoffs + 8 + MARGIN, yoffs + MARGIN, WSIZE, WSIZE,
			MAP_SIZEX*CHIP_SIZEX, MAP_SIZEY*CHIP_SIZEY);
	wchip.font(B_NORMAL_FONT);
	wchip.foreground(wbase.alloc_color(65535, 65535, 65535));
	wchip.background(wbase.alloc_color(0, 0, 0));
	wchip.clear();
	gengine.output(&wchip);

	dhigh.init(&gengine);
	dhigh.place(xoffs + 8 + 240 + 1, yoffs + 8, 64, 18);
	dscore.init(&gengine);
	dscore.place(dhigh.x(), dhigh.y2() + 6, 64, 18);
	wradar.init(&gengine);
	wradar.place(dhigh.x(), yoffs + (SCREEN_HEIGHT-MAP_SIZEY)/2,
			MAP_SIZEX, MAP_SIZEY);
	dships.init(&gengine);
	dships.place(dhigh.x(), wradar.y2() + 6, 38, 18);
	dstage.init(&gengine);
	dstage.place(dhigh.x(), dships.y2() + 6, 38, 18);

	Uint32 dc = wbase.alloc_color(0, 0x3333, 0x6666);

	dhigh.font(B_NORMAL_FONT);
	dhigh.color(dc);
	dhigh.caption("HIGHSCORE");

	dscore.font(B_NORMAL_FONT);
	dscore.color(dc);
	dscore.caption("SCORE");

	dships.font(B_NORMAL_FONT);
	dships.color(dc);
	dships.caption("SHIPS");

	dstage.font(B_NORMAL_FONT);
	dstage.color(dc);
	dstage.caption("STAGE");

	radar.prepare(1);

	if(show_fps)
	{
		dfps.init(&gengine);
		dfps.place(xoffs + wbase.width()-32,
				yoffs + wbase.height()-18, 32, 18);
		dfps.color(wbase.alloc_color(0,0,0));
		dfps.font(B_NORMAL_FONT);
		dfps.caption("FPS");
	}

	gengine.invalidate();
}


void init_audio(prefs_t *p)
{
	audio_master_volume(p->volume);
	if(p->sound)
	{
		if(p->bgm_indexfile[0])
		{
			init_bgm(p->bgm_indexfile);
			p->use_ext_music = 1;
		}
		if(audio_open(p->samplerate, p->latency, p->use_oss) < 0)
			fprintf(stderr, "Couldn't initialize audio;"
					" disabling sound effects.\n");
	}
	audio_master_reverb(p->reverb);
	audio_set_limiter((float)p->threshold / 100.0,
			(float)p->release / 100.0);
	init_sounds();
}


void close_audio()
{
	audio_close();
}


int init_js(prefs_t *p)
{
	/* Activate Joystick sub-sys if we are using it */
	if(p->use_joystick)
	{
		if(SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0)
		{
			fprintf(stderr, "Error setting up joystick!\n");
			return -1;
		}

		if(SDL_NumJoysticks() > 0)
		{
			SDL_JoystickEventState(SDL_ENABLE);
			joystick = SDL_JoystickOpen(0);
			if(!joystick)
			{
				SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
				return -2;
			}
		}
		else
		{
			fprintf(stderr, "No joysticks found!\n");
			joystick = NULL;
			return -3;
		}

	}
	return 0;
}


void close_js()
{
	if(!SDL_WasInit(SDL_INIT_JOYSTICK))
		return;

	if(!joystick)
		return;

	if(SDL_JoystickOpened(0))
		SDL_JoystickClose(joystick);
	joystick = NULL;

	SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
}


void load_config(prefs_t *p)
{
	FILE *f = fmap.fopen(KOBO_CONFIG_DIR "/" KOBO_CONFIG_FILE, "r");
	if(f)
	{
		p->read(f);
		fclose(f);
	}
}


void save_config(prefs_t *p)
{
	FILE *f = fmap.fopen(KOBO_CONFIG_DIR "/" KOBO_CONFIG_FILE, "w");
	if(f)
	{
		p->write(f);
		fclose(f);
	}
}



/*----------------------------------------------------------
	Various handlers...
----------------------------------------------------------*/

void skobo_gfxengine_t::frame()
{
	if(!gsm.current())
	{
		fprintf(stderr, "INTERNAL ERROR: No gamestate!\n");
		exit_game = 1;
		stop();
		return;
	}
	if(exit_game || manage.aborted())
	{
		stop();
		return;
	}

	/*
	 * Process input
	 */
	SDL_Event ev;
	while(SDL_PollEvent(&ev))
	{
		int k;
		switch (ev.type)
		{
		  case SDL_KEYDOWN:
			if(debug_mode)
				switch(ev.key.keysym.sym)
				{
				  case SDLK_F1:
					nibble(0);
					build_screen();
					break;
				  case SDLK_F2:
					nibble(1);
					build_screen();
					break;
				  case SDLK_F3:
					nibble(2);
					build_screen();
					break;
				  case SDLK_F4:
					nibble(3);
					build_screen();
					break;
				  case SDLK_F5:
					nibble(4);
					build_screen();
					break;
				  case SDLK_F9:
					manage.ships = 1;
					myship.destroyed();
					break;
				  default:
					break;
				}
			k = gamecontrol.map(ev.key.keysym.sym);
			gamecontrol.press(k);
			gsm.press(k);
			break;
		  case SDL_KEYUP:
			k = gamecontrol.map(ev.key.keysym.sym);
			gamecontrol.release(k);
			gsm.release(k);
			break;
		  case SDL_QUIT:
			gsm.press(BTN_CLOSE);
			break;
		  case SDL_JOYBUTTONDOWN:
			if(ev.jbutton.button == js_fire)
			{
				gamecontrol.press(BTN_FIRE);
				gsm.press(BTN_FIRE);
			}
			else if(ev.jbutton.button == js_start)
			{
				gamecontrol.press(BTN_START);
				gsm.press(BTN_START);
			}
			break;
		  case SDL_JOYBUTTONUP:
			if(ev.jbutton.button == js_fire)
			{
				gamecontrol.release(BTN_FIRE);
				gsm.release(BTN_FIRE);
			}
			break;
		  case SDL_JOYAXISMOTION:
			// FIXME: We will want to allow these to be
			// redefined, but for now, this works ;-)
			if(ev.jaxis.axis == js_lr)
			{
				if(ev.jaxis.value < -3200)
				{
					gamecontrol.press(BTN_LEFT);
					gsm.press(BTN_LEFT);
				}
				else if(ev.jaxis.value > 3200)
				{
					gamecontrol.press(BTN_RIGHT);
					gsm.press(BTN_RIGHT);
				}
				else
				{
					gamecontrol.release(BTN_LEFT);
					gamecontrol.release(BTN_RIGHT);
					gsm.release(BTN_LEFT);
					gsm.release(BTN_RIGHT);
				}

			}
			else if(ev.jaxis.axis == js_ud)
			{
				if(ev.jaxis.value < -3200)
				{
					gamecontrol.press(BTN_UP);
					gsm.press(BTN_UP);
				}
				else if(ev.jaxis.value > 3200)
				{
					gamecontrol.press(BTN_DOWN);
					gsm.press(BTN_DOWN);
				}
				else
				{
					gamecontrol.release(BTN_UP);
					gamecontrol.release(BTN_DOWN);
					gsm.release(BTN_UP);
					gsm.release(BTN_DOWN);
				}

			}
		}
	}

	/*
	 * Run the current gamestate for one frame
	 */
	gsm.frame();
}


void skobo_gfxengine_t::pre_render()
{
	gsm.pre_render(&wchip);
}


void skobo_gfxengine_t::post_render()
{
	gsm.post_render(&wchip);

	if(!noframe)
		wchip.sprite(0, 0, B_FRAME, 0, 0);

	DBG(if(debug_mode)
	{
		char buf[20];
		snprintf(buf, sizeof(buf), "Obj: %d", gengine.objects_in_use());
		wchip.font(B_NORMAL_FONT);
		wchip.string(160, 5, buf);

		int mx = oscframes;
		if(mx > wchip.width())
			mx = wchip.width();
		int yo = wchip.height() - 40;
		wchip.foreground(wchip.alloc_color(0, 0, 0x9999));
		wchip.fillrect(0, yo, wchip.width(), 1);
		wchip.foreground(wchip.alloc_color(0x9999, 0, 0));
		wchip.fillrect(0, yo - 32, wchip.width(), 1);
		wchip.fillrect(0, yo + 31, wchip.width(), 1);
		wchip.foreground(wchip.alloc_color(0, 0x9999, 0));
		for(int s=0; s<mx; ++s)
			wchip.point(s, (oscbuf[s] >> 11) + yo);

		wchip.foreground(wchip.alloc_color(0xcccc, 0, 0));
		wchip.fillrect(0, yo - 32, 6, attenuation >> 11);
	});

	int nt = (int)SDL_GetTicks();
	int tt = nt - fps_starttime;
	if((tt > 1000) && fps_count)
	{
		float fps = fps_count * 1000.0 / tt;
		::screen.fps(fps);
		fps_count = 0;
		fps_starttime = nt;
		if(show_fps)
		{
			char buf[20];
			snprintf(buf, sizeof(buf), "%.1f", fps);
			dfps.text(buf);
			printf("%s fps\n", buf);
		}
	}
	++fps_count;
}


void kobo_render_highlight(ct_widget_t *wg)
{
	static float ypos = -50;
	static int ot = 0;
	int t = (int)SDL_GetTicks();
	int y;
	int x = -((t>>3) % TUBE_LENGTH);
	int dt = t - ot;
	ot = t;
	if(dt > 1000)
		dt = 100;

	//Spring + friction style velocity component
	if(dt < 500)
		ypos += ((float)wg->y() - ypos) * (float)dt * 0.005;
	else
		ypos = (float)wg->y();

	//Constant velocity component
	float v = (float)dt * 0.1;
	if((float)wg->y() - ypos > v)
		ypos += v;
	else if((float)wg->y() - ypos < -v)
		ypos -= v;
	else
		ypos = (float)wg->y();

	y = (int)(ypos + 0.5) - wchip.y();
	while(x < wg->width())
	{
		if(wg->height() < 12)
			wchip.sprite(x, y, B_SMALL_TUBE, 0);
		else
			wchip.sprite(x, y, B_BIG_TUBE, 0);
		x += TUBE_LENGTH;
	}
}


void close_all(void)
{
	close_js();
	gengine.close();
	close_audio();
	SDL_Quit();
}


RETSIGTYPE breakhandler(int dummy)
{
	if(exit_game_fast)
	{
		atexit(SDL_Quit);
		exit(1);
	}

	exit_game_fast = 1;
	gengine.stop();
}


int main(int argc, char *argv[])
{
	int i;

	atexit(close_all);
	signal(SIGTERM, breakhandler);
	signal(SIGINT, breakhandler);

	setup_dirs(argv[0]);

	if( (argc < 2) || (strcmp("-override", argv[1]) != 0) )
		load_config(&prefs);

	if(prefs.parse(argc, argv) < 0)
	{
		put_usage();
		return 1;
	}

	add_dirs(&prefs);

	for (i = 1; i < argc; i++)
	{
		if(!strcmp("-hiscores", argv[i]))
		{
			scorefile.show_high_scores(1);
			return 0;
		}
		else if(!strcmp("-showcfg", argv[i]))
		{
			printf("Configuration:\n");
			printf("----------------------------------------\n");
			prefs.write(stdout);
			printf("\nPaths:\n");
			printf("----------------------------------------\n");
			fmap.print(stdout, "*");
			return 0;
		}
		else if(!strcmp("-debug", argv[i]))
			debug_mode = 1;
		else if(!strcmp("-fps", argv[i]))
			show_fps = 1;
		else if(!strcmp("-noframe", argv[i]))
			noframe = 1;
	}

	SDL_Init(0);

	if(init_display(&prefs) < 0)
		return 3;

	if(load_graphics(&prefs) < 0)
		return 2;

	ct_engine.render_highlight = kobo_render_highlight;
	build_screen();
	rand_num.init();
	init_audio(&prefs);
	init_js(&prefs);

	gsm.push(&st_intro_1);

	while(1)
	{
		manage.init();
		gengine.run();

		if(exit_game_fast)
			break;

		if(exit_game)
		{
			nibble();
			break;
		}

		if(main_options.status() & OS_RESTART_AUDIO)
		{
			DBG(printf("--- Restarting audio...\n"));
			close_audio();
			init_audio(&prefs);
			DBG(printf("--- Audio restarted.\n"));
		}
		if(main_options.status() & OS_RESTART_VIDEO)
		{
			DBG(printf("--- Restarting video...\n"));
			nibble();
			gengine.close();
			if(init_display(&prefs) < 0)
				return 3;
			DBG(printf("--- Video restarted.\n"));
		}
		if(main_options.status() & (OS_RELOAD_GRAPHICS |
						OS_RESTART_VIDEO))
		{
			DBG(printf("--- Reloading graphics...\n"));
			if(!(main_options.status() & OS_RESTART_VIDEO))
				nibble();
			nibble();
			if(load_graphics(&prefs) < 0)
				return 5;

			build_screen();
			DBG(printf("--- Graphics reloaded.\n"));
		}
		if(main_options.status() & OS_RESTART_ENGINE)
		{
			DBG(printf("--- Restarting engine...\n"));
			gengine.close();
			gengine.scroll_ratio(LAYER_OVERLAY, 0.0, 0.0);
			gengine.scroll_ratio(LAYER_PLAYER, 1.0, 1.0);
			gengine.scroll_ratio(LAYER_ENEMIES, 1.0, 1.0);
			gengine.scroll_ratio(LAYER_BASES, 1.0, 1.0);
			gengine.scroll_ratio(LAYER_GROUND, 1.0, 1.0);
			gengine.scroll_ratio(LAYER_STARS, 0.5, 0.5);
			gengine.wrap(MAP_SIZEX*CHIP_SIZEX, MAP_SIZEY*CHIP_SIZEY);
			if(gengine.open(2048) < 0)
				return 4;
			DBG(printf("--- Engine restarted.\n"));
		}
		if(main_options.status() & OS_RESTART_INPUT)
		{
			close_js();
			init_js(&prefs);
		}
	}

	close_all();

	/*
	 * Seems like we got all the way here without crashing,
	 * so let's save the current configuration! :-)
	 */
	if(prefs.changed)
	{
		save_config(&prefs);
		prefs.changed = 0;
	}

	return 0;
}
