/*
------------------------------------------------------------
   Kobo Deluxe - An enhanced SDL port of XKobo
------------------------------------------------------------
 * Copyright (C) 1995, 1996,  Akira Higuchi
 * Copyright (C) 2001, David Olofson
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "kobo.h"
#include "screen.h"
#include "myship.h"
#include "enemies.h"
#include "gamectl.h"
#include "manage.h"
#include "random.h"
#include "sounds.h"

_myship_state _myship::_state;
int _myship::di;
int _myship::virtx;
int _myship::virty;
int _myship::x;
int _myship::y;
int _myship::lapx;
int _myship::lapy;
int _myship::beamx[BEAM_MAX];
int _myship::beamy[BEAM_MAX];
int _myship::beamdi[BEAM_MAX];
int _myship::beamst[BEAM_MAX];
cs_obj_t *_myship::object;
cs_obj_t *_myship::beam_objects[BEAM_MAX];


_myship::_myship()
{
	object = NULL;
	memset(beam_objects, 0, sizeof(beam_objects));
}


void _myship::state(_myship_state s)
{
	switch (s)
	{
	  case dead:
		if(object)
			gengine.free_obj(object);
		object = NULL;
		break;
	  case normal:
		if(!object)
			object = gengine.get_obj(LAYER_PLAYER);
		if(object)
			cs_obj_show(object);
		break;
	}
	_state = s;
}


void _myship::off()
{
	state(dead);
	int i;
	for(i = 0; i < BEAM_MAX; i++)
		if(beam_objects[i])
		{
			gengine.free_obj(beam_objects[i]);
			beam_objects[i] = NULL;
		}
}


int _myship::init()
{
	x = WORLD_SIZEX >> 1;
	y = (WORLD_SIZEY >> 2) * 3;
	virtx = x - (WSIZE >> 1);
	virty = y - (WSIZE >> 1);
	lapx = 0;
	lapy = 0;
	di = 1;
	state(normal);

	apply_position();

	int i;
	for(i = 0; i < BEAM_MAX; i++)
	{
		beamx[i] = 0;
		beamy[i] = 0;
		beamdi[i] = 0;
		beamst[i] = 0;
		if(beam_objects[i])
			gengine.free_obj(beam_objects[i]);
		beam_objects[i] = NULL;
	}
	return 0;
}


#define BEAMV1           12
#define BEAMV2           (BEAMV1*2/3)

int _myship::move()
{
	int i;
	di = gamecontrol.dir();

	virtx = x - (WSIZE >> 1);
	virty = y - (WSIZE >> 1);
	if(_state == normal)
	{
		int vd, vo;
		if(!prefs.indicator)
		{
			vd = 2;
			vo = 3;
		}
		else if(gamecontrol.movekey_pressed)
		{
			vd = 1;
			vo = 1;
		}
		else
		{
			vd = 0;
			vo = 0;
		}
		switch (di)
		{
		  case 1:
			virty -= vo;
			break;
		  case 2:
			virty -= vd;
			virtx += vd;
			break;
		  case 3:
			virtx += vo;
			break;
		  case 4:
			virtx += vd;
			virty += vd;
			break;
		  case 5:
			virty += vo;
			break;
		  case 6:
			virty += vd;
			virtx -= vd;
			break;
		  case 7:
			virtx -= vo;
			break;
		  case 8:
			virtx -= vd;
			virty -= vd;
			break;
		}
	}
	else if(_state == dead)
	{
		enemies.make(&explosion, x + rand_num.get(6) - 32,
				y + rand_num.get(6) - 32);
		int rn = rand_num.get(8);
		if(rn > 160)
			play_sound(SOUND_EXPL);
		else if(rn > 140)
			play_sound(SOUND_BOMB);
	}

	lapx = 0;
	lapy = 0;
	if(virtx < 0)
	{
		virtx += WORLD_SIZEX;
		lapx = WORLD_SIZEX;
	}
	if(virtx >= WORLD_SIZEX)
	{
		virtx -= WORLD_SIZEX;
		lapx = -WORLD_SIZEX;
	}
	if(virty < 0)
	{
		virty += WORLD_SIZEY;
		lapy = WORLD_SIZEY;
	}
	if(virty >= WORLD_SIZEY)
	{
		virty -= WORLD_SIZEY;
		lapy = -WORLD_SIZEY;
	}
	x = virtx + (WSIZE >> 1);
	y = virty + (WSIZE >> 1);

	if((_state == normal) && gamecontrol.get_shot())
		_myship::shot();

	for(i = 0; i < BEAM_MAX; i++)
	{
		if(!beamst[i])
			continue;
		beamx[i] += lapx;
		beamy[i] += lapy;
		switch (beamdi[i])
		{
		  case 1:
			beamy[i] -= BEAMV1;
			break;
		  case 2:
			beamy[i] -= BEAMV2;
			beamx[i] += BEAMV2;
			break;
		  case 3:
			beamx[i] += BEAMV1;
			break;
		  case 4:
			beamx[i] += BEAMV2;
			beamy[i] += BEAMV2;
			break;
		  case 5:
			beamy[i] += BEAMV1;
			break;
		  case 6:
			beamy[i] += BEAMV2;
			beamx[i] -= BEAMV2;
			break;
		  case 7:
			beamx[i] -= BEAMV1;
			break;
		  case 8:
			beamx[i] -= BEAMV2;
			beamy[i] -= BEAMV2;
			break;
		}
		if((ABS(beamx[i] - x) >= (VIEWLIMIT >> 1) + 16) ||
				(ABS(beamy[i] - y) >= (VIEWLIMIT >> 1) + 16))
		{
			beamst[i] = 0;
			if(beam_objects[i])
				gengine.free_obj(beam_objects[i]);
			beam_objects[i] = NULL;
		}
	}
	return 0;
}


void _myship::destroyed()
{
	if(_state != normal)
		return;
	manage.lost_myship();
	state(dead);
	play_sound(SOUND_EXPL);
	play_sound(SOUND_BOMB);
}


int _myship::hit_structure()
{
	int x1, y1;
	int i, ch;
	for(i = 0; i < BEAM_MAX; i++)
	{
		if(!beamst[i])
			continue;
		x1 = (beamx[i] & (WORLD_SIZEX - 1)) >> 4;
		y1 = (beamy[i] & (WORLD_SIZEY - 1)) >> 4;
		ch = screen.get_chip_number(x1, y1);
		if(!IS_SPACE(ch) && (ch & HIT_MASK))
		{
			play_sound(SOUND_METALLIC);
			beamst[i] = 0;
			if(beam_objects[i])
				gengine.free_obj(beam_objects[i]);
			beam_objects[i] = NULL;
		}
	}
	x1 = (x & (WORLD_SIZEX - 1)) >> 4;
	y1 = (y & (WORLD_SIZEY - 1)) >> 4;
	ch = screen.get_chip_number(x1, y1);
	if(!IS_SPACE(ch) && (ch & HIT_MASK))
	{
		if(prefs.indicator)
			play_sound(SOUND_METALLIC);
		else
			_myship::destroyed();
	}
	return 0;
}


int _myship::hit_beam(int ex, int ey, int hitsize)
{
	int i;
	for(i = 0; i < BEAM_MAX; i++)
	{
		if(beamst[i] == 0)
			continue;
		if(ABS(ex - beamx[i]) >= hitsize)
			continue;
		if(ABS(ey - beamy[i]) >= hitsize)
			continue;
		if(!prefs.cheat_mode)
		{
			beamst[i] = 0;
			if(beam_objects[i])
				gengine.free_obj(beam_objects[i]);
			beam_objects[i] = NULL;
		}
		return 1;
	}
	return 0;
}


int _myship::put()
{
	/* Player */
	apply_position();
	if(object)
		cs_obj_image(object, B_SPRITES, 3 * 16 + di - 1);

	/* Bullets */
	int i;
	for(i = 0; i < BEAM_MAX; i++)
	{
		if(!beam_objects[i])
			continue;
		if(beamst[i])
		{
			beam_objects[i]->point.v.x =
					PIXEL2CS(beamx[i] - (CHIP_SIZEX>>1));
			beam_objects[i]->point.v.y =
					PIXEL2CS(beamy[i] - (CHIP_SIZEY>>1));
		}
	}
#if 0
	if(_state != normal)
		return 0;

	/* Player */
	wchip.copy_from_chip_sprite((di - 1) << (CHIP_SIZEX_LOG2 +
				scale_log2),
			3 << (CHIP_SIZEY_LOG2 + scale_log2),
			CHIP_SIZEX << scale_log2, CHIP_SIZEY << scale_log2,
			(x - CHIP_SIZEX / 2) << scale_log2,
			(y - CHIP_SIZEY / 2) << scale_log2);

	/* Bullets */
	int i;
	for(i = 0; i < BEAM_MAX; i++)
	{
		if(beamst[i])
			wchip.copy_from_chip_sprite((beamdi[i] - 1)
					<< (CHIP_SIZEX_LOG2 + scale_log2),
					2 << (CHIP_SIZEY_LOG2 +
						scale_log2),
					CHIP_SIZEX << scale_log2,
					CHIP_SIZEY << scale_log2,
					(beamx[i] -
						CHIP_SIZEX /
						2) << scale_log2,
					(beamy[i] -
						CHIP_SIZEY /
						2) << scale_log2);
	}

	if((mouse_x == 0) && (mouse_y == 0))
		return 0;

	/* Crosshair */
	wchip.copy_from_chip_sprite(15 << (CHIP_SIZEX_LOG2 + scale_log2),
			5 << (CHIP_SIZEY_LOG2 + scale_log2),
			5 << scale_log2,
			5 << scale_log2,
			(x + mouse_x - 2) << scale_log2,
			(y + mouse_y - 2) << scale_log2);
#endif
	return 0;
}


void _myship::shot_single(int i, int dir)
{
	beamdi[i] = dir;
	beamx[i] = x;
	beamy[i] = y;
	beamst[i] = 1;
	if(!beam_objects[i])
		beam_objects[i] = gengine.get_obj(LAYER_PLAYER);
	if(beam_objects[i])
	{
		beam_objects[i]->point.v.x =
				PIXEL2CS(beamx[i] - (CHIP_SIZEX>>1));
		beam_objects[i]->point.v.y =
				PIXEL2CS(beamy[i] - (CHIP_SIZEY>>1));
		cs_obj_image(beam_objects[i], B_SPRITES, 2 * 16 + beamdi[i] - 1);
		cs_obj_show(beam_objects[i]);
	}
}


int _myship::shot()
{
	int i, j;
	for(i = 0; i < BEAM_MAX && beamst[i]; i++)
		;
	for(j = i + 1; j < BEAM_MAX && beamst[j]; j++)
		;
	if(j >= BEAM_MAX)
		return 1;

	shot_single(i, di);
	shot_single(j, (di > 4) ? (di - 4) : (di + 4) );

	play_sound(SOUND_SHOT);
	return 0;
}


void _myship::set_position(int px, int py)
{
	x = px;
	y = py;
	virtx = x - (WSIZE >> 1);
	virty = y - (WSIZE >> 1);

	if(object)
	{
		apply_position();
		cs_point_force(&object->point);
	}
}


void _myship::apply_position()
{
	if(object)
	{
		object->point.v.x = PIXEL2CS(x - (CHIP_SIZEX>>1));
		object->point.v.y = PIXEL2CS(y - (CHIP_SIZEY>>1));
	}
}
