/*
------------------------------------------------------------
   Kobo Deluxe - An enhanced SDL port of XKobo
------------------------------------------------------------
 * Copyright (C) 2001, David Olofson
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "config.h"
#include "options.h"
#include "gfxengine.h"

main_options_t main_options;

void main_options_t::build()
{
	int i;
	int scale = 1;
	char buf[50];
		form_switch("Use Joystick", &prf->use_joystick,
				OS_RESTART_INPUT);
	form_space();
		form_switch("Fullscreen Display", &prf->fullscreen,
				OS_RESTART_VIDEO);
		form_list("Display Depth", &prf->depth, OS_RESTART_VIDEO);
			form_item("Default", 0);
			form_item("8 bits", 8);
			form_item("15 bits", 15);
			form_item("16 bits", 16);
			form_item("24 bits", 24);
			form_item("32 bits", 32);
		form_list("Display Driver", &prf->videodriver,
				OS_RESTART_VIDEO | OS_REBUILD);
			form_item("SDL 2D", GFX_DRIVER_SDL2D);
#if defined(HAVE_OPENGL) && !defined(GLSDL_OFF)
			form_item("OpenGL Buffer", GFX_DRIVER_OPENGLBUF);
			form_item("Full OpenGL Integer", GFX_DRIVER_OPENGL_INT);
//			form_item("Full OpenGL", GFX_DRIVER_OPENGL);
#else
		prf->videodriver = GFX_DRIVER_SDL2D;
#endif
		switch(prf->videodriver)
		{
		  case GFX_DRIVER_SDL2D:
			form_list("Display Buffer Mode", &prf->buffermode,
					OS_RESTART_VIDEO);
				form_item("Single", GFX_BUFFER_SINGLE);
				form_item("Double", GFX_BUFFER_DOUBLE);
				form_item("SemiTriple", GFX_BUFFER_SEMITRIPLE);
				form_item("Half", GFX_BUFFER_HALF);
			scale = prf->width / 320;
			break;
		  case GFX_DRIVER_OPENGLBUF:
		  case GFX_DRIVER_OPENGL_INT:
		  case GFX_DRIVER_OPENGL:
			form_list("Texture Resolution", &prf->internalres,
					OS_RESTART_VIDEO | OS_REBUILD);
				form_item("320x240", 1);
				form_item("640x480", 2);
				form_item("960x720", 3);
				form_item("1280x1024", 4);
			scale = prf->width / (320*prf->internalres);
			break;
		}
//		form_spin("Scale Factor", &prf->size, 1, 4, TAG_RESTART_VIDEO);
		form_list("Display Size", &prf->width,
				OS_RESTART_VIDEO | OS_REBUILD);
			form_item("320x240", 320);
			form_item("400x300 (OpenGL)", 400);
			form_item("512x384 (OpenGL)", 512);
			form_item("640x480", 640);
			form_item("800x600 (OpenGL)", 800);
			form_item("960x720", 960);
			form_item("1024x768 (OpenGL)", 1024);
			form_item("1152x864 (OpenGL)", 1152);
			form_item("1280x1024", 1280);
		if(scale > 1)
		{
			form_list("Scale Mode", &prf->scalemode,
					OS_RELOAD_GRAPHICS);
				form_item("Nearest", GFX_SCALE_NEAREST);
				form_item("Bilinear", GFX_SCALE_BILINEAR);
				form_item("Bilinear+Oversampling", GFX_SCALE_BILIN_OVER);
			if(2 == scale)
			{
				form_item("Scale2x", GFX_SCALE_SCALE2X);
				form_item("Diamond2x", GFX_SCALE_DIAMOND);
			}
			else if(prf->scalemode >= GFX_SCALE_SCALE2X)
				prf->scalemode = GFX_SCALE_BILINEAR;
		}
	form_space();
		form_spin("Game Speed (ms/frame)", &prf->wait_msec, 10, 100,
				OS_UPDATE_ENGINE);
		form_switch("Motion Interpolation", &prf->filter,
				OS_UPDATE_ENGINE);
	form_space();
		form_switch("Enable Sound", &prf->sound, OS_RESTART_AUDIO | OS_REBUILD);
		if(prf->sound)
		{
			form_list("Sample Rate", &prf->samplerate, OS_RESTART_AUDIO);
				form_item("8000", 8000);
				form_item("11025", 11025);
				form_item("16000", 16000);
				form_item("22050", 22050);
				form_item("32000", 32000);
				form_item("44100", 44100);
				form_item("48000", 48000);
			form_list("Sound Volume", &prf->volume, OS_UPDATE_AUDIO);
				for(i = 25; i <= 200; i += 25)
				{
					snprintf(buf, sizeof(buf), "%d", i);
					form_item(buf, i);
				}
			form_spin("Sound Latency", &prf->latency, 1, 200,
					OS_RESTART_AUDIO);
#ifdef OSS_SUPPORT
			form_switch("Use OSS Sound Driver", &prf->use_oss,
					OS_RESTART_AUDIO);
#endif
			form_list("Reverb Level", &prf->reverb, OS_UPDATE_AUDIO);
				form_item("OFF", 0);
				for(i = 25; i <= 200; i += 25)
				{
					snprintf(buf, sizeof(buf), "%d", i);
					form_item(buf, i);
				}
			form_list("Limiter Speed", &prf->release, OS_UPDATE_AUDIO);
				form_item("Slow", 40);
				form_item("Normal", 100);
				form_item("Fast", 250);
				form_item("Aggressive", 500);
			form_list("Limiter Gain", &prf->threshold, OS_UPDATE_AUDIO);
				form_item("None", 100);
				form_item("Low", 90);
				form_item("Normal", 80);
				form_item("High", 60);
				form_item("Pumping", 30);
				form_item("Extreme", 10);
		}
	form_space();
		form_button("ACCEPT", OS_CLOSE);
		form_button("CANCEL", OS_CANCEL);
}
