/*
------------------------------------------------------------
   Kobo Deluxe - An enhanced SDL port of XKobo
------------------------------------------------------------
 * Copyright (C) 1995, 1996, Akira Higuchi
 * Copyright (C) 2001, David Olofson
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>

#include "kobo.h"
#include "screen.h"
#include "manage.h"
#include "options.h"
#include "enemies.h"
#include "myship.h"
//#include "gamectl.h"
#include "radar.h"
#include "scenes.h"
#include "config.h"
#include "random.h"
#include "version.h"

int _screen::scene_max;
int _screen::scene_num;
int _screen::level;
int _screen::generate_count;
_map _screen::map;
int _screen::pixel_f0 = -1;
int _screen::pixel_f1 = -1;
int _screen::pixel_f2 = -1;
int _screen::pixel_s[8];
int _screen::show_title = 0;
float _screen::_fps = 30;
int _screen::scroller_speed = SCROLLER_SPEED;

/* Translate "map code" into tile index */
int _screen::translate(int n)
{
	if(IS_SPACE(n))
		return (n & 0xff) + 64;

	int x = n;
	int y = 7;
	if(n & HARD)
	{
		y = 6;
		if(n & U_MASK)
			x = 8;
		else if(n & R_MASK)
			x = 9;
		else if(n & D_MASK)
			x = 10;
		else if(n & L_MASK)
			x = 11;
	}
	else if(n & CORE)
	{
		y = 6;
		x = 15;
	}
	return y * 16 + x - 104;
}


void _screen::init()
{
	scene_max = 0;
	while(scene[scene_max].ratio != -1)
		scene_max++;
	if(pixel_f0 < 0)
	{
		pixel_f0 = wbase.alloc_color(200 << 8, 255 << 8, 255 << 8);
		pixel_f1 = wbase.alloc_color(140 << 8, 200 << 8, 200 << 8);
		pixel_f2 = wbase.alloc_color(100 << 8, 200 << 8, 100 << 8);
		int i;
		for(i = 0; i < 8; i++)
		{
			pixel_s[i] = wbase.alloc_color(40000,
					15000 + i * 4000,
					50000 - i * 5000);
		}
	}
}


/*
FIXME: This should be replaced with some nice scrolling stuff.
*/
void _screen::title()
{
	wchip.foreground(pixel_f1);
	wchip.sprite((wchip.width() - 208)/2, 40, B_LOGO, 0);
	wchip.font(B_NORMAL_FONT);
	wchip.center(115, "v" KOBO_VERSION);
	if(prefs.cheat_mode && (manage.count & 0x10))
		wchip.center(130, "CHEAT MODE");
	wchip.foreground(pixel_f2);
}


void _screen::highscores()
{
	wchip.font(B_BIG_FONT);
	wchip.center(20, "HALL OF FAME");
	wchip.font(B_NORMAL_FONT);
	wchip.center(100, "(Not yet implemented)");
}


void _screen::credits()
{
	wchip.font(B_BIG_FONT);
	wchip.center(20, "KOBO DELUXE");

	wchip.font(B_NORMAL_FONT);
	wchip.center(45, "XKobo (The Original Game)");
	wchip.center(55, "AKIRA HIGUCHI");

	wchip.center(75, "New SDL Engine, Aditional GFX");
	wchip.center(85, "DAVID OLOFSON");

	wchip.center(105, "Sound & Original SFX Engine");
	wchip.center(115, "MASANAO IZUMO");

	wchip.center(135, "Additional Credits & Thanks");
	wchip.center(145, "See scroller below");
}


void _screen::scroller()
{
	/*
	 * Adjust scroller speed according to
	 * frame rate, for readability.
	 */
	if(_fps < 50)
	{
		if(scroller_speed > SCROLLER_SPEED / 2)
			scroller_speed -= 3;
		else
			scroller_speed = SCROLLER_SPEED / 2;
	}
	else
	{
		if(scroller_speed < SCROLLER_SPEED)
			scroller_speed += 1;
		else
			scroller_speed = SCROLLER_SPEED;
	}

	static const char scrolltext[] =
			"Welcome to KOBO DELUXE, an enhanced version of "
			"Akira Higuchi's fabulous X-Windows game XKOBO. "
			"      "
			"This version uses SDL (http://www.libsdl.org/) "
			"for graphics, sound and input, and should be "
			"easy to port to most reasonable platforms. "
			"      "
			"KOBO DELUXE is already known to build and run on:"
			"  -  "
			"GNU/Linux/x86  -  "
			"FreeBSD/x86  -  "
			"Windows 95/98/2000  -  "
			"Sun Solaris  -  "
			"Mac OS X  -  "
			"                        "
			"Additional Credits "
			"(See README file for details):"
			"      Max Horn (Mac OS X)"
			"      Tsuyoshi Iguchi (FreeBSD)"
			"      G. Low (Solaris)"
			"      Gerry Jo \"Trick\" Jellestad (Testing & Ideas)"
			"      Other Members of the SDL Mailing List"
			"      My Girlfriend, Ingela"
			"                        "
			"Additional Thanks from Akira Higuchi Go To:"
			"      Bruce Cheng"
			"      Christoph Lameter"
			"      Davide Rossi"
			"      Eduard Martinescu"
			"      Elan Feingold"
			"      Helmut Hoenig"
			"      Jeff Epler"
			"      Joe Ramey"
			"      Joey Hess"
			"      Michael Sterrett"
			"      Mihail Iotov"
			"      Shoichi Nakayama"
			"      Thomas Marsh"
			"      Torsten Wolnik"
			"                        "
			"Now, if you were around the so called "
			"\"Demo Scene\" back in the days of the "
			"glorious C64, Amiga and Atari ST machines, "
			"you're about to see a familiar term... :-)"
			"                        "
			"         <WRAP>         ";
	static const char *stp = scrolltext;
	static int pos = -PIXEL2CS(SCREEN_WIDTH);
	static int t = 0;
	static int offs = 0;
	char buf[2] = {0, 0};
	int nt = (int)SDL_GetTicks();
	int dt = nt - t;
	t = nt;
	if(dt > 100)
		dt = 100;

#if 1
	static int fdt = 0;
	fdt += ((dt<<8) - fdt) >> 2;
#else
	int fdt = dt<<8;
#endif
	pos += (fdt * PIXEL2CS(scroller_speed) / 1000) >> 8;
	wchip.font(B_BIG_FONT);
	wchip.string_fxp(-pos + offs, PIXEL2CS(180), stp);

	/*
	 * Chop away characters at the left edge
	 */
	buf[0] = *stp;
	int cw = wchip.textwidth(buf);
	if(CS2PIXEL(-pos + offs) < -cw)
	{
		offs += PIXEL2CS(cw);
		++stp;
		if(*stp == 0)
		{
			pos = -PIXEL2CS(SCREEN_WIDTH);
			stp = scrolltext;
			offs = 0;
		}
	}
}


void _screen::usage()
{
	int t = SDL_GetTicks();
	float ft = t * 0.001;
	int y = PIXEL2CS(180) + (int)floor(PIXEL2CS(5)*sin(ft * 6));

	if(t % 3000 < 2500)
	{
		wchip.font(B_BIG_FONT);
		switch((t/3000) % 6)
		{
		  case 0:
			wchip.center_fxp(y, "---- INSTRUCTIONS ----");
			break;
		  case 1:
			wchip.center_fxp(y, "Press SPACE to Start Game!");
			break;
		  case 2:
			wchip.center_fxp(y, "Use +/- to Select Start Stage");
			break;
		  case 3:
			wchip.center_fxp(y, "Hit Backspace to Rebuild Stage");
			break;
		  case 4:
			wchip.center_fxp(y, "Hit RETURN for Options");
			break;
		  case 5:
			wchip.center_fxp(y, "(Don't) Hit ESC to Quit");
			break;
		}
	}
}


void _screen::init_scene(int sc)
{
	map.init();
	if(sc < 0)
	{
		/*
		 * Intro mode
		 */
//		scene_num = -1;
		show_title = 1;
		myship.off();
		enemies.off();
		if(sc == INTRO_SCENE)
		{
			/*
			 * Plain intro - no map
			 */
			radar.prepare();	/* Clear radar */
			scene_num = 19 % scene_max;
			level = 19 / scene_max;
		}
		else
		{
			/*
			 * Map selection - show current map
			 */
			scene_num = -(sc+1) % scene_max;
			level = -(sc+1) / scene_max;
		}
	}
	else
	{
		/*
		 * In-game mode
		 */
		gengine.period(0);
		show_title = 0;
		scene_num = sc % scene_max;
		level = sc / scene_max;
	}
//printf("scene: %d, level: %d\n", scene_num, level);

	_scene *s = &scene[scene_num];
	int i;
	for(i = 0; i < s->base_max; i++)
		map.make_maze(s->base[i].x, s->base[i].y, s->base[i].h,
				s->base[i].v);
	map.convert(s->ratio);

	/* No radar map in intro mode! */
	if(sc != INTRO_SCENE)
		radar.prepare();

	generate_count = 0;
}


int _screen::prepare()
{
	if(scene_num < 0)
		return 0;
	_scene *s = &scene[scene_num];
	int i, j;
	int count_core = 0;
	int c = 0;

	int interval_1, interval_2;
	interval_1 = (s->ek1_interval) >> level;
	interval_2 = (s->ek2_interval) >> level;
	if(interval_1 < 4)
		interval_1 = 4;
	if(interval_2 < 4)
		interval_2 = 4;
	enemies.set_ekind_to_generate(s->ek1, interval_1, s->ek2,
			interval_2);

	wchip.clear();
	for(i = 0; i < MAP_SIZEX; i++)
		for(j = 0; j < MAP_SIZEY; j++)
		{
			int m = map.pos(i, j);
			if((m == U_MASK) || (m == R_MASK) || (m == D_MASK)
					|| (m == L_MASK))
			{
				enemies.make(&cannon, i * 16 + 8,
						j * 16 + 8);
				c++;
			}
			else if(m == CORE)
			{
				enemies.make(&core, i * 16 + 8,
						j * 16 + 8);
				count_core++;
				c++;
			}
		}

#if 0
	int x, y;
	for(i = 0; i < 8; i++)
	{
		wchip.foreground(pixel_s[i]);
		for(j = 0; j < 2000; j++)
		{
			x = rand_num.get(WORLD_SIZEX_LOG2);
			y = rand_num.get(WORLD_SIZEY_LOG2);
			if(IS_SPACE(map.pos(x >> 4, y >> 4)))
				wchip.fillrectangle(x, y, 1, 1);
//                wchip.fillrectangle_back(x, y, 1, 1);
		}
	}
#endif

	myship.set_position(s->startx << 4, s->starty << 4);

	radar.prepare();
	return count_core;
}


void _screen::generate_fixed_enemies()
{
	static int sint[16] =
			{ 0, 12, 23, 30, 32, 30, 23, 12, 0, -12, -23, -30,
				-32, -30, -23, -12 };
	static int cost[16] =
			{ 32, 30, 23, 12, 0, -12, -23, -30, -32, -30, -23,
				-12, 0, 12, 23, 30 };
	_scene *s = &scene[scene_num];
	if(generate_count < s->enemy_max)
	{
		int j;
		for(j = 0; j < s->enemy[generate_count].num; j++)
		{
			int sp = s->enemy[generate_count].speed;
			int x, y, h, v, t;
			x = rand_num.get() % (WORLD_SIZEX - VIEWLIMIT * 2);
			y = rand_num.get() % (WORLD_SIZEY - VIEWLIMIT * 2);
			x -= (WORLD_SIZEX / 2 - VIEWLIMIT);
			y -= (WORLD_SIZEY / 2 - VIEWLIMIT);
			if(x < 0)
				x -= VIEWLIMIT;
			else
				x += VIEWLIMIT;
			if(y < 0)
				y -= VIEWLIMIT;
			else
				y += VIEWLIMIT;
			x += myship.get_x();
			y += myship.get_y();

			t = rand_num.get(4);
			h = PIXEL2CS(sp * sint[t]) / 64;
			v = PIXEL2CS(sp * cost[t]) / 64;
			enemies.make(s->enemy[generate_count].kind, x, y,
					h, v);
		}
		generate_count++;
	}
	if(generate_count >= s->enemy_max)
		generate_count = 0;
}


int _screen::get_chip_number(int x, int y)
{
	return map.pos(x, y);
}


void _screen::set_chip_number(int x, int y, int n)
{
	if(n == 0)
	{
		map.clearpos(x, y);
		radar.erase(x, y);
	}
	else
		map.pos(x, y) = n;

}


void _screen::render_background(window_t *win)
{
	if(!win)
		return;

	int vx, vy, xo, yo, x, y;
	int mx, my;
	vx = gengine.xoffs(LAYER_BASES);
	vy = gengine.yoffs(LAYER_BASES);

	/* 
	 * Start exactly at the top-left corner
	 * of the tile visible in the top-left
	 * corner of the display window.
	 */
	xo = vx & (PIXEL2CS(CHIP_SIZEX) - 1);
	yo = vy & (PIXEL2CS(CHIP_SIZEY) - 1);
	mx = CS2PIXEL(vx >> CHIP_SIZEX_LOG2);
	my = CS2PIXEL(vy >> CHIP_SIZEY_LOG2);

	for(y = 0; y < (WSIZE >> CHIP_SIZEY_LOG2) + 2; ++y)
		for(x = 0; x < (WSIZE >> CHIP_SIZEX_LOG2) + 2; ++x)
		{
			int n = map.pos(mx + x, my + y);
#if 0
			if(IS_SPACE(n))
			{
				win->foreground(win->alloc_color(0,0,0));
				win->fillrect_fxp(PIXEL2CS(x<<CHIP_SIZEX_LOG2) - xo,
						PIXEL2CS(y<<CHIP_SIZEX_LOG2) - yo,
						PIXEL2CS(CHIP_SIZEX),
						PIXEL2CS(CHIP_SIZEY));
			}
			else
#endif
			win->sprite_fxp(PIXEL2CS(x<<CHIP_SIZEX_LOG2) - xo,
					PIXEL2CS(y<<CHIP_SIZEX_LOG2) - yo,
					B_TILES, translate(n));
		}
}


void _screen::fps(float f)
{
	_fps = f;
}
